
/*
 * tkImgXPM.c --
 *
 * A photo image file handler for XPM files. 
 *
 * This needs libXpm 4.3
 *
 * Copyright (c) 1996 by Alfredo Kengi Kojima 
 *
 * Adapted from tkImgBmap.c
 * 
 * Copyright (c) 1994 The Regents of the University of California.
 * Copyright (c) 1994-1996 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 * 
 * and tkImgGIF.c
 * 
 * Reed Wade (wade@cs.utk.edu), University of Tennessee
 *
 * Copyright (c) 1995-1996 Sun Microsystems, Inc.
 *
 */

#ifdef ENABLE_XPM

#include "tkInt.h"
#include "tkPort.h"
#include <X11/xpm.h>

/*
 * The type record for XPM images:
 */
static int      ImgXPMStringMatch _ANSI_ARGS_((char *string,
			char *formatString, int *widthPtr, int *heightPtr));

static int 	ImgXPMStringRead _ANSI_ARGS_((Tcl_Interp *interp,
			char *string, char *formatString, 
			Tk_PhotoHandle imageHandle,
			int destX, int destY, int width, int height, 
			int srcX, int srcY));

static int      ImgXPMFileMatch _ANSI_ARGS_((FILE *f, char *fileName,
			char *formatString, int *widthPtr, int *heightPtr));

static int      ImgXPMFileRead  _ANSI_ARGS_((Tcl_Interp *interp,
		    	FILE *f, char *fileName, char *formatString,
		    	Tk_PhotoHandle imageHandle, int destX, int destY,
		    	int width, int height, int srcX, int srcY));


Tk_PhotoImageFormat tkImgFmtXPM = {
	"XPM",			/* name */
	ImgXPMFileMatch,  	/* fileMatchProc */
	ImgXPMStringMatch,     	/* stringMatchProc */
	ImgXPMFileRead,    	/* fileReadProc */
	ImgXPMStringRead,     	/* stringReadProc */
	NULL,           	/* fileWriteProc */
	NULL,	           	/* stringWriteProc */
};


#define MAXCOLORMAPSIZE		256
#define CM_RED			0
#define CM_GREEN		1
#define CM_BLUE			2

/*
 * Prototypes for procedures used only locally in this file:
 */

static int		MakeColortable _ANSI_ARGS_((Tcl_Interp *interp, 
			    XpmImage image, 
			    unsigned char   cmap[3][MAXCOLORMAPSIZE], 
			    int *transparent));
static int		ReadImage _ANSI_ARGS_((Tcl_Interp *interp,
			    char *imagePtr, XpmImage image, int len, int height,
			    unsigned char cmap[3][MAXCOLORMAPSIZE],
			    int transparent));

/*
 *----------------------------------------------------------------------
 *
 * ImgXPMFileMatch --
 *
 *  This procedure is invoked by the photo image type to see if
 *  a file contains image data in XPM format.
 *
 * Results:
 *  The return value is 1 if the first characters in file f look
 *  like XPM data, and 0 otherwise. 
 *  The XPM file must contain the keyword XPM in the first line
 *
 * Side effects:
 *  None
 *
 *----------------------------------------------------------------------
 */

static int
ImgXPMFileMatch(f, fileName, formatString, widthPtr, heightPtr)
    FILE *f;			/* The image file, open for reading. */
    char *fileName;		/* The name of the image file. */
    char *formatString;		/* User-specified format string, or NULL. */
    int *widthPtr, *heightPtr;	/* The dimensions of the image are
				 * returned here if the file is a valid
				 * XPM file. */
{
    XpmImage image;
    
    if (XpmReadFileToXpmImage(fileName, &image, (XpmInfo*)NULL)!=XpmSuccess) {
	return 0;
    }
    *widthPtr = image.width;
    *heightPtr = image.height;
    XpmFreeXpmImage(&image);
    return 1;
}

/*
 *----------------------------------------------------------------------
 *
 * ImgXPMFileRead --
 *
 *	This procedure is called by the photo image type to read
 *	XPM format data from a file and write it into a given
 *	photo image.
 *
 * Results:
 *	A standard TCL completion code.  If TCL_ERROR is returned
 *	then an error message is left in interp->result.
 *
 * Side effects:
 *	New data is added to the image given by imageHandle.
 *
 *----------------------------------------------------------------------
 */

static int
ImgXPMFileRead(interp, f, fileName, formatString, imageHandle, destX, destY,
	width, height, srcX, srcY)
    Tcl_Interp *interp;		/* Interpreter to use for reporting errors. */
    FILE *f;			/* The image file, open for reading. */
    char *fileName;		/* The name of the image file. */
    char *formatString;		/* User-specified format string, or NULL. */
    Tk_PhotoHandle imageHandle;	/* The photo image to write into. */
    int destX, destY;		/* Coordinates of top-left pixel in
				 * photo image to be written to. */
    int width, height;		/* Dimensions of block of photo image to
				 * be written to. */
    int srcX, srcY;		/* Coordinates of top-left pixel to be used
				 * in image being read. */
{
    int fileWidth, fileHeight;
    int nBytes;
    Tk_PhotoImageBlock block;
    unsigned char colorMap[3][MAXCOLORMAPSIZE];
    int transparent = -1;
    XpmImage image;
    
    if (XpmReadFileToXpmImage(fileName, &image, (XpmInfo *)NULL)!=XpmSuccess) {
	Tcl_AppendResult(interp, "error reading XPM file \"", fileName,
		"\"", (char *) NULL);
	return TCL_ERROR;
    }
    fileWidth = image.width;
    fileHeight = image.height;
    if ((fileWidth <= 0) || (fileHeight <= 0)) {
	Tcl_AppendResult(interp, "XPM image file \"", fileName,
		"\" has dimension(s) <= 0", (char *) NULL);
	XpmFreeXpmImage(&image);
	return TCL_ERROR;
    }

    if ((srcX + width) > fileWidth) {
	width = fileWidth - srcX;
    }
    if ((srcY + height) > fileHeight) {
	height = fileHeight - srcY;
    }
    if ((width <= 0) || (height <= 0)
	    || (srcX >= fileWidth) || (srcY >= fileHeight)) {
	XpmFreeXpmImage(&image);
	return TCL_OK;
    }

    if (MakeColortable(interp, image, colorMap, &transparent) != TCL_OK) {
	goto error;
    }
    Tk_PhotoExpand(imageHandle, destX + width, destY + height);

    block.width = fileWidth;
    block.height = fileHeight;
    block.pixelSize = 3;
    block.pitch = 3 * fileWidth;
    block.offset[0] = 0;
    block.offset[1] = 1;
    block.offset[2] = 2;
    nBytes = fileHeight * block.pitch;
    block.pixelPtr = (unsigned char *) ckalloc((unsigned) nBytes);

    if (ReadImage(interp, (char *) block.pixelPtr, image, fileWidth,
		  fileHeight, colorMap, 0) != TCL_OK) {
	goto error;
    }

    Tk_PhotoPutBlock(imageHandle, &block, destX, destY, fileWidth, fileHeight);
    ckfree((char *) block.pixelPtr);
    XpmFreeXpmImage(&image);
    return TCL_OK;

    error:
    ckfree((char *) block.pixelPtr);
    XpmFreeXpmImage(&image);    
    return TCL_ERROR;

}


/*
 *----------------------------------------------------------------------
 *
 * ImgXPMStringMatch --
 *
 *  This procedure is invoked by the photo image type to see if
 *  a string contains image data in XPM format.
 *
 * Results:
 *  The return value is 1 if the first characters in the string look
 *  like XPM data, and 0 otherwise. 
 *
 * Side effects:
 *  None
 *
 *----------------------------------------------------------------------
 */
static int      
ImgXPMStringMatch (string, formatString, widthPtr, heightPtr)
char *string;
char *formatString;
int *widthPtr;
int *heightPtr;
{
    XpmImage image;
    
    if (XpmCreateXpmImageFromData(&string, &image, (XpmInfo *)NULL)
	!=XpmSuccess) {
	return 0;
    }
    *widthPtr = image.width;
    *heightPtr = image.height;
    XpmFreeXpmImage(&image);
    return 1;
}

/*
 *----------------------------------------------------------------------
 *
 * ImgXPMStringRead --
 *
 *	This procedure is called by the photo image type to read
 *	XPM format data from a string and write it into a given
 *	photo image.
 *
 * Results:
 *	A standard TCL completion code.  If TCL_ERROR is returned
 *	then an error message is left in interp->result.
 *
 * Side effects:
 *	New data is added to the image given by imageHandle.
 *
 *----------------------------------------------------------------------
 */
static int 	
ImgXPMStringRead (interp, string, formatString, imageHandle, destX, destY, 
		  width, height, srcX, srcY)
Tcl_Interp *interp;
char *string;
char *formatString;
Tk_PhotoHandle imageHandle;
int destX, destY, width, height, srcX, srcY;
{
    int nBytes;
    Tk_PhotoImageBlock block;
    unsigned char colorMap[3][MAXCOLORMAPSIZE];
    int transparent = -1;
    XpmImage image;

    
    if (XpmCreateXpmImageFromData(&string, &image, (XpmInfo *)NULL)
	!=XpmSuccess) {
	return 0;
    }

    if ((image.width <= 0) || (image.height <= 0)) {
	Tcl_AppendResult(interp, "XPM image data has dimension(s) <= 0", 
			 (char *) NULL);
	XpmFreeXpmImage(&image);
	return TCL_ERROR;
    }

    if ((srcX + width) > image.width) {
	width = image.width - srcX;
    }
    if ((srcY + height) > image.height) {
	height = image.height - srcY;
    }
    if ((width <= 0) || (height <= 0)
	    || (srcX >= image.width) || (srcY >= image.height)) {
	XpmFreeXpmImage(&image);
	return TCL_OK;
    }

    if (MakeColortable(interp, image, colorMap, &transparent) != TCL_OK) {
	goto error;
    }
    Tk_PhotoExpand(imageHandle, destX + width, destY + height);

    block.width = image.width;
    block.height = image.height;
    block.pixelSize = 3;
    block.pitch = 3 * image.width;
    block.offset[0] = 0;
    block.offset[1] = 1;
    block.offset[2] = 2;
    nBytes = image.height * block.pitch;
    block.pixelPtr = (unsigned char *) ckalloc((unsigned) nBytes);

    if (ReadImage(interp, (char *) block.pixelPtr, image, image.width,
		  image.height, colorMap, 0) != TCL_OK) {
	goto error;
    }

    Tk_PhotoPutBlock(imageHandle, &block, destX, destY, image.width, 
		     image.height);
    ckfree((char *) block.pixelPtr);
    XpmFreeXpmImage(&image);
    return TCL_OK;

    error:
    ckfree((char *) block.pixelPtr);
    XpmFreeXpmImage(&image);    
    return TCL_ERROR;
}


/*
 * Private 
 */

/*
 *------------------------------------------------------------------ 
 * 
 * MakeColortable --
 * 	Converts a XpmColor colortable to a indexed RGB colortable
 * 
 *------------------------------------------------------------------ 
 */
static int
MakeColortable(interp, image, cmap, transparent)
Tcl_Interp *interp;
XpmImage image;
unsigned char   cmap[3][MAXCOLORMAPSIZE];
int *transparent;
{
    int i;
    
    if (image.colorTable==NULL) 
      return TCL_ERROR;
    
    for (i=0; i<image.ncolors; i++) {
	XColor *colorPtr;
	
	if (strncmp(image.colorTable[i].c_color,"None",4)==0) {
	    *transparent = i;
	    cmap[CM_RED][i]=0xbf;
	    cmap[CM_GREEN][i]=0xbf;
	    cmap[CM_BLUE][i]=0xbf;
	    continue;
	}
	colorPtr = Tk_GetColor(interp, Tk_MainWindow(interp), 
			       Tk_GetUid(image.colorTable[i].c_color));
	if (colorPtr) {
	    cmap[CM_RED][i] = colorPtr->red >> 8;
	    cmap[CM_GREEN][i] = colorPtr->green >> 8;
	    cmap[CM_BLUE][i] = colorPtr->blue >> 8;
	    Tk_FreeColor(colorPtr);
	} else {
	    cmap[CM_RED][i]=0xbf;
	    cmap[CM_GREEN][i]=0xbf;
	    cmap[CM_BLUE][i]=0xbf;
	}
    }
    return TCL_OK;
}

/*
 *------------------------------------------------------------------ 
 * 
 * ReadImage --
 * 	Converts a XpmImage image to a RGB image
 * 
 *------------------------------------------------------------------ 
 */
static int
ReadImage(interp, imagePtr, image, len, height, cmap, transparent)
Tcl_Interp *interp;
char 	*imagePtr;
XpmImage image;
int len, height;
unsigned char   cmap[3][MAXCOLORMAPSIZE];
int transparent;
{
    int     v;
    int     xpos = 0, ypos = 0;
    char 	*colStr;
    unsigned int offset;
    
    if (transparent != -1) {
	colStr = Tcl_GetVar(interp, "TRANSPARENT_XPM_COLOR", 0L);
	if (colStr != NULL) {
	    XColor *colorPtr;
	    colorPtr = Tk_GetColor(interp, Tk_MainWindow(interp), 
				   Tk_GetUid(colStr));
	    if (colorPtr) {
		cmap[CM_RED][transparent] = colorPtr->red >> 8;
		cmap[CM_GREEN][transparent] = colorPtr->green >> 8;
		cmap[CM_BLUE][transparent] = colorPtr->blue >> 8;
		Tk_FreeColor(colorPtr);
	    }
	}
    }
    
    offset = 0;
    for(ypos=0; ypos<height; ypos++) {
	for(xpos=0; xpos<len; xpos++) {
	    v = image.data[offset];
	    imagePtr[ (xpos*3)  +  (ypos *len*3)] = cmap[CM_RED][v];
	    imagePtr[ (xpos*3)  +  (ypos *len*3) +1] = cmap[CM_GREEN][v];
	    imagePtr[ (xpos*3)  +  (ypos *len*3) +2] = cmap[CM_BLUE][v];
	    offset++;
	}
    }
    return TCL_OK;    
}
#endif
